import { IDIDDocument } from '@ew-did-registry/did-resolver-interface';
import { Claims } from '../claims';
import { IClaimsVerifier } from '../interface';
import { IPublicClaim } from '../models';
export declare class ClaimsVerifier extends Claims implements IClaimsVerifier {
    /**
     * Verifies integrity of the claim, the claim is issued by the user
     *  delegate and the authenticity of the issuer's signature
     *
     * @example
     * ```typescript
     * import { ClaimsVerifier } from '@ew-did-registry/claims';
     * import { Keys } from '@ew-did-registry/keys';
     *
     * const keys = new Keys();
     * const claims = new ClaimsVerifier(verifier);
     * const verified = claims.verifyPublicProof(issuedToken);
     * ```
     * @param { string } token containing proof data
     * @returns { Promise<void> } whether the proof was succesfull
     * @throws if the proof failed
     */
    verifyPublicProof(claimUrl: string, { holderDoc, issuerDoc, }?: {
        holderDoc?: IDIDDocument;
        issuerDoc?: IDIDDocument;
    }): Promise<IPublicClaim>;
    /**
     * Checks issuer signature on issued token and user signature on proof token
     * and verifies that proof and private data mathches to each other
     *
     * @example
     * ```typescript
     * import { ClaimsVerifier } from '@ew-did-registry/claims';
     * import { Keys } from '@ew-did-registry/keys';
     *
     * const keys = new Keys();
     * const claims = new ClaimsVerifier(verifier);
     * const verified = claims.verifyPrivateProof(proofToken);
     * ```
     * @param { string } proofToken contains proof data
     * @param { string } privateToken contains private data
     * @returns { Promise<void> } whether the proof was succesfull
     * @throws if the proof failed
     */
    verifyPrivateProof(proofToken: string): Promise<void>;
}
